//CG - 22/23 - efolioB - bezier4 Helio Sousa 2000027

import * as THREE from 'https://cdn.skypack.dev/three@0.132.2/build/three.module.js';

// Formula curva de bezier de grau 4 
// f(t)= (1-t)^4*C0 + 4(1-t)^3*t*C1 + 6(1-t)^2*t^2C2 + 4(1-t)t^3*C3 + t^4*C4 , 0 <= t >= 1

// Cálculo dos termos C0, C1, C2, C3 e C4 da curva de bezier de grau 4
function termC0 (c0,t){
    return Math.pow((1-t),4)*c0;
}
function termC1 (c1,t){
    return 4*Math.pow((1-t),3)*t*c1;
}
function termC2 (c2,t){
    return 6*Math.pow((1-t),2)*Math.pow(t,2)*c2;
}
function termC3 (c3,t){
    return 4*(1-t)*Math.pow(t,3)*c3;
}
function termC4 (c4,t){
    return Math.pow(t,4)*c4;
}

// Soma os termos da curva de bezier de grau 4 num dado instante t
function addTerms (c0,c1,c2,c3,c4,t){
    return termC0(c0,t)+termC1(c1,t)+termC2(c2,t)+termC3(c3,t)+termC4(c4,t);
}

/**
 * Cálculo da curva de bezier de grau 4
 * @param {*} pontos de controlo da curva de bezier de grau 4 em THREE.Vector3 e um dado instante t
 * @returns a posição da curva de bezier de grau 4 no instante t em THREE.Vector3
 */

function bezier4 ({c0,c1,c2,c3,c4,t}){
    let curveB4 = new THREE.Vector3();
    curveB4.x=addTerms(c0.x,c1.x,c2.x,c3.x,c4.x,t); // preenche a coordenada x da curva no instante t
    curveB4.y=addTerms(c0.y,c1.y,c2.y,c3.y,c4.y,t); // preenche a coordenada y da curva no instante t
    curveB4.z=addTerms(c0.z,c1.z,c2.z,c3.z,c4.z,t); // preenche a coordenada z da curva no instante t
    return curveB4;                                 // retorna a posição da curva no instante t
}

export {bezier4};
