// Autor: Hélio Sousa
// Trabalho de Computação Gráfica - Efólio A - 2022-2023
// bezier4
// PARTE II - Interface Gráfico (2 valores)

// Importação de módulos
import * as THREE from 'https://cdn.skypack.dev/three@0.132.2';
import { OrbitControls } from 'https://cdn.skypack.dev/three@0.132.2/examples/jsm/controls/OrbitControls.js'
import { bezier4 } from '../bezier4.mjs';

// Inicialização da função principal
document.addEventListener('DOMContentLoaded', start);

// Variaveis globais
let sizePixel = 1;                      // tamanho do pixel(e tile) irão manter-se sempre a 1
let sizeRasterDisplay = 10;             // tamanho da grelha inicial
let rows = sizeRasterDisplay;           // número de colunas do display raster para respeitar o enunciado, é um tabuleiro de lados iguais
let cols = sizeRasterDisplay;           // número de linhas do display raster para respeitar o enunciado, é um tabuleiro de lados iguais
let colorSq1 = 0xff8966;                // consultado em https://imagecolorpicker.com/en -> cor (0,0)
let colorSq2 = 0x8d88b5;                // consultado em https://imagecolorpicker.com/en -> cor (0,1)
let balls = [];                         // array que irá receber as bolas 
let ballsLines=[];                      // array que irá receber as linhas das bolas
let raycast_L1 = new THREE.Raycaster(); // cria um raycaster para detetar interseções com o tabuleiro na layer1
raycast_L1.layers.set(1);               // define a layer do raycaster para a layer1
let mouse = new THREE.Vector2();        // cria um vector2 para receber as posições do rato devolvidas por onMouseMove 
let mouseIntersects=new THREE.Vector2();// Valor que irá receber o ponto de interseção do ray com o tabuleiro   
let verticalPressed = false;            // Booleano que indica se a tecla W está pressionada

let selBall = {
    selected: false,    // Booleano que indica se alguma bola está selecionada
    num: 0              // número da bola selecionada
}; 

// Definir parametros da cena
// renderer 
let renderer = new THREE.WebGLRenderer();               // cria uma instancia de renderer
renderer.setSize(window.innerWidth, window.innerHeight);
document.body.appendChild(renderer.domElement);
// Scene
let scene = new THREE.Scene();
scene.background = new THREE.Color( 0xffffff );         // muda a cor do background da cena para branco
// Camera
let camera = new THREE.PerspectiveCamera(90, window.innerWidth / window.innerHeight, 1, 1000);
scene.add(camera); // confirmar se não existe um add a mais da camera

// orbitControls -> https://threejs.org/docs/#examples/en/controls/OrbitControls
let orbitControls = new OrbitControls(camera, renderer.domElement); //chamada dos orbit controls para o utilizador interagir com a cena
camera.position.set(0, 0, sizeRasterDisplay+1); // colocar a camera numa posição inicial que permita ver todo o tabuleiro
orbitControls.update(); // atualizar os orbit controls
camera.lookAt( 0, 0, 0 );                       // faz a camera olhar para o centro da scene

// definir luz ambiente
let lightAmbient = new THREE.AmbientLight(0xffffff, 1); // branco = 0xffffff
scene.add(lightAmbient);

// FUNÇÃO MAIN + ATUALIZAR FRAMES
/**
 * Inicia a animação, função principal chamada no DOMContentLoaded
 */
function start() {
    displayRaster();
    createAxisHelper();
    createBalls();
    createLineBalls()
    requestAnimationFrame(animate); // chamada da função animate
}
/**
 * Atualiza a cena com as entradas do utilizador
 */
function animate() {
    renderer.render(scene, camera);
    requestAnimationFrame(animate);  // chamada recursiva da função animate
}

// EVENTOS
/**
 * Evento que recolhe os movimentos do rato e atualiza a posição da bola selecionada
 */
document.body.addEventListener('mousemove', ev=> {  
    mouse.x = (ev.clientX / window.innerWidth) * 2 - 1;                 // atualiza as coordenadas do eixo X do rato em relação á janela
    mouse.y = - (ev.clientY / window.innerHeight) * 2 + 1;              // atualiza as coordenadas do eixo Y do rato em relação á janela
    if (selBall.selected && !verticalPressed) {                         // Caso esteja alguma bola selecionada e a tecla W ou S não esteja pressionada
        let C = balls[selBall.num];                                     // coloca em C a bola selecionada
        raycast_L1.setFromCamera(mouse, camera);                        // inicializa o raycaster com as posições do rato em relação á camara
        let intersects = raycast_L1.intersectObjects (scene.children);  // verifica se o raycaster intersecta algum objeto da cena na layer 1
        if (intersects.length > 0) {                                    // Se existir alguma interseção
            C.position.x = intersects[0].point.x;                       // atualiza a coordenada x da bola selecionada com a do ponto de interseção
            C.position.y = intersects[0].point.y;                       // atualiza a coordenada x da bola selecionada com a do ponto de interseção
            mouseIntersects = intersects[0].point;                      // atualiza o ponto de interseção do rato com o tabuleiro
            updateCoordenates();                                        // atualiza as coordenadas das bolas para mostrar
                updateLineBalls(selBall.num);                              // atualiza a linha que liga as bolas selecionadas
        }
    }
});
// Evento que realizar operações ao clicar numa tecla
document.body.addEventListener('keydown', ev=> {
        let num = parseInt(ev.key);  // coloca o valor da tecla pressionada em num
        if (selBall.selected == false && ( num == 1 || num == 2 || num == 3 || num == 4 || num == 5)) { // Caso não esteja nenhuma bola selecionada e a tecla pressionada seja 1,2,3,4 ou 5
            ballSelected(num - 1); // Chama a função ballSelected com o valor da tecla pressionada - 1 (para que o valor seja 0,1,2,3 ou 4)
        }    
        if (selBall.selected == true) { // Caso esteja alguma bola selecionada
            if (ev.code === 'Space'){     // Caso a tecla pressionada seja o espaço
                ballDeselected();         // Chama a função ballDeselected
                verticalPressed = false;  // e coloca verticalPressed a false
            }
            if (ev.code == 'KeyW') {      // Caso a tecla pressionada seja w adiciona 0.1 unidades a coordenada z
                verticalPressed = true;
                balls[selBall.num].position.z = balls[selBall.num].position.z + 0.1;
                    updateLineBalls(selBall.num);                              // atualiza a linha que liga as bolas selecionadas
                updateCoordenates();      // e atualiza a informação das coordenadas no ecrã
            }   
            if (ev.code == 'KeyS') {      // Caso a tecla pressionada seja s subtrai 0.1 unidades a coordenada z
                verticalPressed = true;
                balls[selBall.num].position.z = balls[selBall.num].position.z - 0.1;
                    updateLineBalls(selBall.num);                              // atualiza a linha que liga as bolas selecionadas
                updateCoordenates();      // e atualiza a informação das coordenadas no ecrã
            }
        }   
        if (ev.code == 'KeyX') {          // Caso a tecla pressionada seja x
            let bezierCurve4 = createBezier4();   // cria uma curva de bezier com as bolas selecionadas
            bezierCurve4.layers.enable(4);
            scene.add(bezierCurve4);      // adiciona a curva de bezier à cena
        }
        if (ev.code === 'Backspace') { // Caso a tecla pressionada seja o backspace
            resetScene();  // Chama a função resetScene com a cena base
        }
});

// FUNÇÕES AUXILIARES
/**
 * Função que cria os 3 eixos com tamanho igual ao do display raster fonte: https://threejs.org/docs/#api/en/helpers/AxesHelper
 */
function createAxisHelper() {
    let axes = new THREE.AxesHelper(sizeRasterDisplay);
    axes.layers.enable(2);
    scene.add(axes);
}

/**
 * Função que cria as esferas que representam as bolas
 */
function createBalls() {
    balls.push(createSphere("c0","yellow"));// Cria a esfera amarela
    balls.push(createSphere("c1","orange"));//    "    "   laranja
    balls.push(createSphere("c2","red"));   //   "    "   vermelha
    balls.push(createSphere("c3","green")); //  "    "   verde
    balls.push(createSphere("c4","blue"));  //   "    "   azul
    balls.every (ball => scene.add(ball));  // Insere as esferas na cena
}

/**
 * Função que cria uma esfera com um determinado raio, cor e posição
 * @param {*} i número da esfera
 */
function ballSelected(i) {
    selBall.selected = true;                // a bola está selecionada
    selBall.num = i;                        // coloca o numero da bola selecionada em selBall.num
    balls[i].material.transparent = false;  // desativa a transparencia da bola selecionada
    showCoordenates(balls[i]);              // mostra as coordenadas da bola selecionada
}

/**
 * Função que deseleciona a bola selecionada
 */
function ballDeselected() {
    selBall.selected = false;                       // boladeixa de estar selecionada
    balls[selBall.num].material.transparent = true; // volta a colocar a transparencia a bola selecionada

    removeCoordenates();                            // remove o div com as coordenadas da bola selecionada
}

/**
 * Função que cria um DIV com a informação das coordenadas da bola selecionada
 */
function showCoordenates() {
    let C = balls[selBall.num];                                 // esfera selecionada
    let coordDiv = document.createElement('div');               // cria um div
    document.body.appendChild(coordDiv);                        // adiciona o div ao body
    coordDiv.innerHTML = '<span id="coord"> Coordenadas de C'   // insere o texto no div
    + selBall.num + ' (' + C.position.x.toFixed(2)
    + ',' + C.position.y.toFixed(2) + ',' 
    + C.position.z.toFixed(2) + ')</span>';
}

/**
 * Função que atualiza as coordenadas da bola selecionada no ecrã
 */
function updateCoordenates() {
    let C = balls[selBall.num];             // esfera selecionada
    if (document.getElementById('coord')) { // Caso exista o div com as coordenadas
        document.getElementById('coord')    // atualiza as coordenadas com as coordenadas da bola selecionada
        .innerHTML = 'Coordenadas de C'     
        + selBall.num + ' ('                
        + mouseIntersects.x.toFixed(2)      
        + ',' + mouseIntersects.y.toFixed(2)
        + ',' + C.position.z.toFixed(2)
        + ')';
    }
}

/**
 * Função que remove o DIV
 */
function removeCoordenates() {
    if (document.getElementById('coord')) {                 // Caso exista o div com as coordenadas
        let coordDiv = document.getElementById('coord');    // seleciona o div
        coordDiv.parentNode.removeChild(coordDiv);          // remove o div
    }
}

/**
 * Função que vai criar o display raster, recebe como arg o tamanho da grelha, 
 * o tamanho de cada pixel (lado do quadrado) a cor1 e a cor2
 * adaptado de https://physevangertis.medium.com/loops-in-javascript-make-checkerboards-cf9c17308072
 */
function displayRaster() {
    let color;
    for (let x = -rows; x < rows; x++) {        //ciclo em x de -tamanho da grelha até +tamanho da grelha
        for (let y = -cols; y < cols; y++) {    //ciclo em y de -tamanho da grelha até +tamanho da grelha
            if ((y % 2 == 0 && x % 2 == 0) ||   // se y e x forem pares
                (y % 2 != 0 && x % 2 != 0)) {   // ou se y e x forem impares
                color = colorSq1;               // então a cor é a cor1
            } else {
                color = colorSq2;               // se não a cor é a cor2
            }
            let pixel = new createPixel(sizePixel, color, x, y);// coloca em pixel o pixel(mesh) criado nas posições (x,y) com a cor apropriada
            pixel.layers.enable(1);                             // ativa a camada 1 para o pixel
            scene.add(pixel);                                   // adiciona o pixel à scene
        }
    }
}
/** 
 * Função que cria um pixel para ser colocado no display raster
 * @param {*} side          // tamanho do lado do pixel
 * @param {*} pixelColor    // cor do pixel
 * @param {*} x             // posição x do pixel (+0.5 para ficar no centro do pixel)
 * @param {*} y             // posição y do pixel (+0.5 para ficar no centro do pixel)
 * @returns                 // retorna o pixel criado
 */
function createPixel(side, pixelColor, x, y) {
    const geometry = new THREE.PlaneGeometry(side, side);   // criação da geometria de um plano com o tamanho unitário, sizel pixe = 1
    let material = new THREE.MeshBasicMaterial({            // definição do material 
        opacity: 0.4,                                       // opacidade
        transparent: true,                                  // transparencia
        color: pixelColor,                                  // cor
        side: THREE.DoubleSide,                             // duplo lado
    });
    let pixel = new THREE.Mesh(geometry, material);         // sqr vai ser a mesh da boxgeometry com o material
    pixel.position.set(x+0.5,y+0.5,0);                      // coloca o pixel na posição x,y do display raster, com z=0
    return pixel;                                           // retorna o pixel
}

/**
 * Função que cria uma esfera com as coordenadas iniciais aleatórias
 * @param {*} name nome do objeto 3D
 * @param {*} color cor do objeto 3D
 * @returns uma esfera com as coordenadas iniciais aleatórias
 */
function createSphere (name, color){
    let geometry = new THREE.SphereGeometry( 0.5, 32, 32 ); // esfera com raio 0.5
    let material = new THREE.MeshBasicMaterial( 
        { color: color,                                     // cor da bola
         opacity: 0.7,                                      // opacidade de 0.5
        transparent: true } );                              // transparente
    let sp = new THREE.Mesh( geometry, material );          // cria o material
    let maxRandom=sizeRasterDisplay;                        // valor máximo para a posição inicial aleatória
    sp.position.x=(Math.random()*maxRandom*2-maxRandom);    // posição inicial aleatória
    sp.position.y=(Math.random()*maxRandom*2-maxRandom);    // posição inicial aleatória
    sp.position.z=0;                                        // como pedido no enunciado, posição z=0 no inicio
    sp.userData.id=sp.id;                                   // id da bola guardada no userData
    sp.name = name;                                         // nome da bola guardada no userData
    sp.userData.color = color;                              // cor da bola guardada no userData
    return sp;                                              // retorna a esfera (bola)
}

/**
 * Função que cria um objeto 3D de tubegeometry com o caminho
 * definido por uma curva bezier quatrica
 * @returns retorna um tubegeometry com a curva bezier quatrica
 */
function createBezier4(){
    class bezierCurvePath extends THREE.Curve { // Classe extendida de THREE.Curve fonte: https://threejs.org/docs/#api/en/geometries/TubeGeometry
        constructor( scale = 1 ) {              // construtor da classe
            super();                            // super classe esta classe estende a classe THREE.Curve
            this.scale = scale;                 // escala
        }
        getPoint(t) {                           // t vai de 0 a 1
            return bezier4({
                c0:balls[0].position,           // posição da bola 0
                c1:balls[1].position,           // posição da bola 1
                c2:balls[2].position,           // posição da bola 2
                c3:balls[3].position,           // posição da bola 3
                c4:balls[4].position,           // posição da bola 4
                t:t                             // valor de t
            }); 
        }
    }
    let path = new bezierCurvePath();           // Criação da path
    let geometry = new THREE.TubeGeometry(      // Criação da geometria do tubo
        path, 20, 0.25, 8, false);              // path, numero de segmentos, raio, numero de rotações, fechado
    let material = new THREE.LineBasicMaterial({// Material do tubo
        color: "cyan",                          // cor
        side: THREE.DoubleSide,                 // duplo lado
        transparent: true,                      // transparencia
        opacity: 0.8,                           // opacidade
    });
    let bezierCurve = new THREE.Mesh (          // criação do objeto
        geometry, material );             
    return bezierCurve;                         // retorna o objeto
}

/**
 * Função que cria um objeto 3D de tubegeometry com o caminho
 */
function createLineBalls() {                                // criação das linhas
    for(let i=0;i<balls.length;i++){                        // para cada bola
        const lineMaterial = new THREE.LineBasicMaterial(   // material da linha
            {color: balls[i].userData.color});              // cor da linha igual à cor da bola, guardada previamente em userData
        let lineGeometry=new THREE.BufferGeometry()         // geometria da linha
        .setFromPoints(0,0,0,0,0,0);                        // criação da linha com 2 pontos a zero
        let line=new THREE.Line(lineGeometry,lineMaterial); // criação da linha
        line.layers.enable(3);                              // layer da linha
        ballsLines.push(line);                              // adiciona a linha ao array de linhas
        scene.add(line);                                    // adiciona a linha à scene
    }
}

/**
 * Função que atualiza a posição das linhas
 */
function updateLineBalls(i){                                 // atualização das linhas
    ballsLines[i].geometry.setFromPoints(               // atualiza a posição da linha
        [new THREE.Vector3(balls[i].position.x,balls[i] // nova posição inicial da linha
            .position.y,0),                              
        new THREE.Vector3(balls[i].position.x,balls[i]  // nova posição final da linha
            .position.y,balls[i].position.z),]           
    )
}

// VALE A PENA CRIAR UM REMOVE LINE ONMOUSE MOVE EVENT!!!


/**
 * Função que faz reset à cena
 */
function resetScene() {
    scene.remove.apply(scene, scene.children);      // remove todos os objetos da cena
    balls = [];                                     // array de bolas
    ballsLines=[];                                  // array de linhas
    createAxisHelper();                             // criação do eixo
    displayRaster();                                // criação do tabuleiro
    createBalls();                                  // criação das bolas
    createLineBalls();                              // criação das linhas
    camera.position.set(0, 0, sizeRasterDisplay+1); // posição da
    camera.lookAt(0, 0, 0);                         // posição para onde a camera está a olhar
}